/*
    webmail -- simple pop3/smtp web e-mail cgi
    Copyright (C) 2000-2004 Dan Cahill

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* FUNCTION LIST
 * int  printhex(const char *format, ...);
 * int  printht(const char *format, ...);
 * void printline(char *msgtext);
 * int  EncodeBase64(char *src, int srclen);
 * int  DecodeBase64(char *src);
 * char *DecodeRFC2047(char *src);
 * int  DecodeHTML(short int reply, char *src, char *ctype, short int crlf);
 * int  DecodeQP(char *src);
 * int  DecodeText(short int reply, char *src);
 */
#include "main.h"

static const char Base64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

int printhex(const char *format, ...)
{
	char *hex="0123456789ABCDEF";
	unsigned char buffer[1024];
	int offset=0;
	va_list ap;

	va_start(ap, format);
	vsnprintf(buffer, sizeof(buffer)-1, format, ap);
	va_end(ap);
	while (buffer[offset]) {
		if ((buffer[offset]>32)&&(buffer[offset]<128)&&(buffer[offset]!='<')&&(buffer[offset]!='>')) {
			printf("%c", buffer[offset]);
		} else {
			printf("%%%c%c", hex[(unsigned int)buffer[offset]/16], hex[(unsigned int)buffer[offset]&15]);
		}
		offset++;
	}
	return 0;
}

int printht(const char *format, ...)
{
	unsigned char buffer[1024];
	int offset=0;
	va_list ap;

	va_start(ap, format);
	vsnprintf(buffer, sizeof(buffer)-1, format, ap);
	va_end(ap);
	while (buffer[offset]) {
		if (buffer[offset]=='<') {
			printf("&lt;");
//		} else if (buffer[offset]=='>') {
//			printf("&gt;");
		} else if (buffer[offset]=='&') {
			printf("&amp;");
		} else {
			printf("%c", buffer[offset]);
		}
		offset++;
	}
	return 0;
}

void printline(short int reply, char *msgtext)
{
	char *pmsgbody;
	char *ptemp;
	char line[120];
	unsigned int width=78;

	pmsgbody=msgtext;
	if (reply) width=76;
	while (strlen(pmsgbody)>width) {
		memset(line, 0, sizeof(line));
		snprintf(line, width+1, "%s", pmsgbody);
		line[width]='\0';
		if (reply) printf("> ");
		if (strchr(line, '\r')||strchr(line, '\n')) {
			if ((ptemp=strchr(line, '\r'))!=NULL) *ptemp='\0';
			if ((ptemp=strchr(line, '\n'))!=NULL) *ptemp='\0';
			printht("%s\r\n", line);
			pmsgbody+=strlen(line);
			if (*pmsgbody=='\r') pmsgbody++;
			if (*pmsgbody=='\n') pmsgbody++;
		} else if ((ptemp=strrchr(line, ' '))!=NULL) {
			*ptemp='\0';
			printht("%s\r\n", line);
			pmsgbody+=strlen(line)+1;
		} else {
			printht("%s", line);
			pmsgbody+=strlen(line);
		}
	}
	if (reply) printf("> ");
	printht("%s\r\n", pmsgbody);
}

char *decodedate(char *src)
{
	int dim[12]={ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
	char *months[]={
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	};
	char *ptemp=src;
	int i;
	int stat;
	int day, month, year;
	int hours, minutes, seconds;
	int tzoffset;
	time_t unixdate=-1;
	static char newdate[40];

	memset(newdate, 0, sizeof(newdate));
	while ((*ptemp)&&(!isdigit(*ptemp))) ptemp++;
	if(!isdigit(*ptemp)) return "";
	if (*ptemp=='0') ptemp++;
	day=atoi(ptemp);
	while ((*ptemp)&&(!isalpha(*ptemp))) ptemp++;
	stat=-1;
	for (i=0;i<12;i++) {
		if (strncasecmp(ptemp, months[i], 3)==0) stat=i;
	}
	if (stat==-1) return "";
	month=stat;
	while ((*ptemp)&&(!isdigit(*ptemp))) ptemp++;
	year=atoi(ptemp);
	for (i=1970;i<year;i++) {
		unixdate+=365;
		if ((i/4.0f)==(int)(i/4)) {
			if ((i/400.0f)==(int)(i/400)) {
				unixdate++;
			} else if ((i/100.0f)!=(int)(i/100)) {
				unixdate++;
			}
		}
	}
	for (i=0;i<month;i++) {
		unixdate+=dim[i];
		if (i!=1) continue;
		if ((year/4.0f)==(int)(year/4)) {
			if ((year/400.0f)==(int)(year/400)) {
				unixdate++;
			} else if ((year/100.0f)!=(int)(year/100)) {
				unixdate++;
			}
		}
	}
	unixdate+=day;
	if (unixdate<1) return "";
	unixdate*=86400;
	while ((*ptemp)&&(isdigit(*ptemp))) ptemp++;
	while ((*ptemp)&&(!isdigit(*ptemp))) ptemp++;
	hours=atoi(ptemp);
	while ((*ptemp)&&(isdigit(*ptemp))) ptemp++;
	while ((*ptemp)&&(!isdigit(*ptemp))) ptemp++;
	minutes=atoi(ptemp);
	while ((*ptemp)&&(isdigit(*ptemp))) ptemp++;
	while ((*ptemp)&&(!isdigit(*ptemp))) ptemp++;
	seconds=atoi(ptemp);
	while ((*ptemp)&&(isdigit(*ptemp))) ptemp++;
	while ((*ptemp)&&(!isdigit(*ptemp))&&(*ptemp!='-')) ptemp++;
	tzoffset=atoi(ptemp);
	unixdate+=hours*3600;
	unixdate+=minutes*60;
	unixdate+=seconds;
	unixdate-=(int)(tzoffset/100)*3600;
	strftime(newdate, sizeof(newdate)-1, "%b %d, %Y %I:%M%p", localtime(&unixdate));
	newdate[18]=tolower(newdate[18]);
	newdate[19]=tolower(newdate[19]);
	return newdate;
}

char *DecodeRFC2047(char *src)
{
	static char dest[1024];
	char *pos;
	int destidx;
	int state;
	int ch=0;
	int szdest;

	memset(dest, 0, sizeof(dest));
	szdest=sizeof(dest)-1;
	destidx=0;
	while (*src) {
		if (strncmp(src, "=?", 2)!=0) {
			dest[destidx++]=*src++;
			continue;
		}
		while ((*src)&&(*src!='?')) src++;
		if (*src=='\0') return dest;
		src++;
		// we don't try to make sense of the charset
		while ((*src)&&(*src!='?')) src++;
		if (*src=='\0') return dest;
		src++;
		if (tolower(*src)=='q') {
			while ((*src)&&(*src!='?')) src++;
			if (*src=='\0') return dest;
			src++;
			while ((ch=*src++)!='\0') {
				if (ch=='?') break;
				if (ch=='=') {
					if (isxdigit(src[0])&&isxdigit(src[1])) {
						dest[destidx++]=(char)IntFromHex(src);
						src+=2;
					}
				} else {
					dest[destidx++]=ch;
				}
				if (destidx>=szdest) return dest;
			}
		} else if (tolower(*src)=='b') {
			while ((*src)&&(*src!='?')) src++;
			if (*src=='\0') return dest;
			src++;
			state=0;
			while ((ch=*src++)!='\0') {
				if (isspace(ch)) continue;
				if (ch=='=') break;
				pos=strchr(Base64, ch);
				if (pos==0) break;
				switch (state) {
				case 0:
					if (dest) {
						if (destidx>=szdest) return dest;
						dest[destidx]=(pos-Base64)<<2;
					}
					state=1;
					break;
				case 1:
					if (dest) {
						if (destidx+1>=szdest) return dest;
						dest[destidx]|=(pos-Base64)>>4;
						dest[destidx+1]=((pos-Base64)&0x0f)<<4;
					}
					destidx++;
					state=2;
					break;
				case 2:
					if (dest) {
						if (destidx+1>=szdest) return dest;
						dest[destidx]|=(pos-Base64)>>2;
						dest[destidx+1]=((pos-Base64)&0x03)<<6;
					}
					destidx++;
					state=3;
					break;
				case 3:
					if (dest) {
						if (destidx>=szdest) return dest;
						dest[destidx]|=(pos-Base64);
					}
					destidx++;
					state=0;
					break;
				}
			}
		}
		if (ch=='=') {
			while ((*src)&&(*src!='?')) src++;
		}
		if (*src=='\0') return dest;
		src++;
		if (*src=='=') src++;
		ch=0;
	}
	return dest;
}

int DecodeHTML(short int reply, char *src, char *ctype, short int crlf)
{
	char dest[1024];
	char *destidx;

	if (strncasecmp(ctype, "text/html", 9)!=0) return 0;
	memset(dest, 0, sizeof(dest));
	destidx=dest;
	if (reply) printf("> ");
	while ((*src)&&(strlen(dest)<sizeof(dest))) {
		if (strncasecmp(src, "<SCRIPT", 7)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<EMBED", 6)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<HEAD", 5)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<HTML", 5)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<IFRAME", 7)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<LINK", 5)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<OBJECT", 7)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (strncasecmp(src, "<PARAM", 6)==0) {
			while ((*src)&&(*src!='>')) src++;
			if (*src) src++;
		} else if (reply) {
			if ((strncasecmp(src, "<BR>", 4)==0)||(strncasecmp(src, "<P>", 3)==0)||(strncasecmp(src, "<TR", 3)==0)) {
				while ((*src)&&(*src!='>')) src++;
				if (*src) src++;
				printf("%s\r\n", dest);
				if (reply) printf("> ");
				memset(dest, 0, sizeof(dest));
				destidx=dest;
			} else if (strncasecmp(src, "<LI>", 4)==0) {
				while ((*src)&&(*src!='>')) src++;
				if (*src) src++;
				printf("%s\r\n", dest);
				if (reply) printf("> ");
				printf("* ");
				memset(dest, 0, sizeof(dest));
				destidx=dest;
			} else if (strncasecmp(src, "<", 1)==0) {
				while ((*src)&&(*src!='>')) src++;
				if (*src) src++;
			} else {
				*destidx++=*src++;
			}
		} else {
			*destidx++=*src++;
		}
	}
	if (strlen(dest)>0) {
//		if (reply) printf("> ");
		printf("%s", dest);
		if (crlf) printf("\r\n");
	}
	return (strlen(dest));
}

int DecodeQP(short int reply, char *src, char *ctype)
{
	char dest[1024];
	char *destidx;
	short int crlf=1;

	memset(dest, 0, sizeof(dest));
	destidx=dest;
	while ((*src)&&(strlen(dest)<sizeof(dest))) {
		if (*src=='=') {
			src++;
			if (*src==0) {
				crlf=0;
			} else if (isxdigit(src[0])&&isxdigit(src[1])) {
				*destidx++=(char)IntFromHex(src);
				src+=2;
			}
		} else if ((*src=='\r')||(*src=='\n')) {
			src++;
		} else {
			*destidx++=*src++;
		}
	}
	if (strncasecmp(ctype, "text/html", 9)==0) {
		DecodeHTML(reply, dest, ctype, crlf);
	} else {
		if (reply) printf("> ");
		printht("%s", dest);
		if (src[strlen(src)-1]!='=') printf("\r\n");
	}
	return (strlen(dest));
}

int DecodeText(short int reply, char *src)
{
	char dest[1024];
	char *destidx;

	memset(dest, 0, sizeof(dest));
	destidx=dest;
	while ((*src)&&(strlen(dest)<sizeof(dest)-1)) {
		if (*src=='\r') {
			*destidx++=*src++;
			if (*src!='\n') *destidx++='\n';
		} else {
			*destidx++=*src++;
		}
	}
	if (reply) printf("> ");
	printf("%s\r\n", dest);
	return (strlen(dest));
}

int EncodeBase64(char *src, int srclen)
{
	unsigned char a, b, c, d, *cp;
	int dst, i, enclen, remlen, linelen;

	cp=src;
	dst=0;
	linelen=0;
	enclen=srclen/3;
	remlen=srclen-3*enclen;
	for (i=0;i<enclen;i++) {
		a=(cp[0]>>2);
	        b=(cp[0]<<4)&0x30;
		b|=(cp[1]>>4);
		c=(cp[1]<<2)&0x3c;
		c|=(cp[2]>>6);
		d=cp[2]&0x3f;
		cp+=3;
		wmprintf("%c%c%c%c", Base64[a], Base64[b], Base64[c], Base64[d]);
		dst+=4;
		linelen+=4;
		if (linelen>=76) {
			wmprintf("\r\n");
			linelen=0;
		}
	}
	if (remlen==0) {
		wmprintf("\r\n");
	} else if (remlen==1) {
		a=(cp[0]>>2);
		b=(cp[0]<<4)&0x30;
		wmprintf("%c%c==\r\n", Base64[a], Base64[b]);
		dst+=4;
	} else if (remlen==2) {
		a=(cp[0]>>2);
		b=(cp[0]<<4)&0x30;
		b|=(cp[1]>>4);
		c=(cp[1]<<2)&0x3c;
		wmprintf("%c%c%c=\r\n", Base64[a], Base64[b], Base64[c]);
		dst+=4;
	}
	return dst;
}

int DecodeBase64(char *src, char *ctype)
{
	char dest[1024];
	int destidx, state, ch;
	int szdest;
	char *pos;

	memset(dest, 0, sizeof(dest));
	szdest=sizeof(dest)-1;
	state=0;
	destidx=0;
	while ((ch=*src++)!='\0') {
		if (isspace(ch)) continue;
		if (ch=='=') break;
		pos=strchr(Base64, ch);
		if (pos==0) return (-1);
		switch (state) {
			case 0:
				if (dest) {
					if (destidx>=szdest) return (-1);
					dest[destidx]=(pos-Base64)<<2;
				}
				state=1;
				break;
			case 1:
				if (dest) {
					if (destidx+1>=szdest) return (-1);
					dest[destidx]|=(pos-Base64)>>4;
					dest[destidx+1]=((pos-Base64)&0x0f)<<4;
				}
				destidx++;
				state=2;
				break;
			case 2:
				if (dest) {
					if (destidx+1>=szdest) return (-1);
					dest[destidx]|=(pos-Base64)>>2;
					dest[destidx+1]=((pos-Base64)&0x03)<<6;
				}
				destidx++;
				state=3;
				break;
			case 3:
				if (dest) {
					if (destidx>=szdest) return (-1);
					dest[destidx]|=(pos-Base64);
				}
				destidx++;
				state=0;
				break;
		}
	}
	if (strncasecmp(ctype, "text/html", 9)==0) {
		DecodeHTML(0, dest, ctype, 0);
	} else if (strncasecmp(ctype, "text/plain", 10)==0) {
		printf("%s", dest);
	} else {
		fwrite(dest, sizeof(char), destidx, stdout);
	}
	return (szdest);
}

char *EncodeBase64string(char *src)
{
	unsigned char a, b, c, d, *cp;
	static char dest[1024];
	int dst, i, enclen, remlen;
	int srclen=strlen(src);

	memset(dest, 0, sizeof(dest));
	cp=src;
	dst=0;
	enclen=srclen/3;
	remlen=srclen-3*enclen;
	for (i=0;i<enclen;i++) {
		a=(cp[0]>>2);
	        b=(cp[0]<<4)&0x30;
		b|=(cp[1]>>4);
		c=(cp[1]<<2)&0x3c;
		c|=(cp[2]>>6);
		d=cp[2]&0x3f;
		cp+=3;
		dest[dst+0]=Base64[a];
		dest[dst+1]=Base64[b];
		dest[dst+2]=Base64[c];
		dest[dst+3]=Base64[d];
		dst+=4;
		if (dst>=1020) return dest;
	}
	if (remlen==1) {
		a=(cp[0]>>2);
		b=(cp[0]<<4)&0x30;
		dest[dst+0]=Base64[a];
		dest[dst+1]=Base64[b];
		dest[dst+2]='=';
		dest[dst+3]='=';
	} else if (remlen==2) {
		a=(cp[0]>>2);
		b=(cp[0]<<4)&0x30;
		b|=(cp[1]>>4);
		c=(cp[1]<<2)&0x3c;
		dest[dst+0]=Base64[a];
		dest[dst+1]=Base64[b];
		dest[dst+2]=Base64[c];
		dest[dst+3]='=';
	}
	return dest;
}

char *DecodeBase64string(char *src)
{
	static char dest[1024];
	int destidx, state, ch;
	int szdest;
	char *pos;

	memset(dest, 0, sizeof(dest));
	szdest=sizeof(dest)-1;
	state=0;
	destidx=0;
	while ((ch=*src++)!='\0') {
		if (isspace(ch)) continue;
		if (ch=='=') break;
		pos=strchr(Base64, ch);
		if (pos==0) return "";
		switch (state) {
			case 0:
				if (dest) {
					if (destidx>=szdest) return dest;
					dest[destidx]=(pos-Base64)<<2;
				}
				state=1;
				break;
			case 1:
				if (dest) {
					if (destidx+1>=szdest) return dest;
					dest[destidx]|=(pos-Base64)>>4;
					dest[destidx+1]=((pos-Base64)&0x0f)<<4;
				}
				destidx++;
				state=2;
				break;
			case 2:
				if (dest) {
					if (destidx+1>=szdest) return dest;
					dest[destidx]|=(pos-Base64)>>2;
					dest[destidx+1]=((pos-Base64)&0x03)<<6;
				}
				destidx++;
				state=3;
				break;
			case 3:
				if (dest) {
					if (destidx>=szdest) return dest;
					dest[destidx]|=(pos-Base64);
				}
				destidx++;
				state=0;
				break;
		}
	}
	return dest;
}

/*
 * EncodeRFC2047(), jis2euc() and euc2jis()
 * done by MAEDA, Go <g_maeda@mars.dti.ne.jp>
 */
char *EncodeRFC2047(char *src)
{
	static char dest[1024];
	char jissrc[1024], encoded[1024];
	char *cp, *cq;
	int i, srclen, enclen, remlen, a, b, c, d;
	
#ifdef SUPPORT_JIS
	euc2jis(src, jissrc, sizeof(jissrc));
#else
	snprintf(jissrc, sizeof(jissrc)-1, "%s", src);
#endif
	cp=jissrc;
	cq=encoded;
	srclen = strlen(jissrc);
	enclen=srclen/3;
	remlen=srclen-3*enclen;
	for (i=0;i<enclen;i++) {
		a=(cp[0]>>2);
	        b=(cp[0]<<4)&0x30;
		b|=(cp[1]>>4);
		c=(cp[1]<<2)&0x3c;
		c|=(cp[2]>>6);
		d=cp[2]&0x3f;
		cp+=3;
		*cq++ = Base64[a];
		*cq++ = Base64[b];
		*cq++ = Base64[c];
		*cq++ = Base64[d];
	}
	if (remlen==1) {
		a=(cp[0]>>2);
		b=(cp[0]<<4)&0x30;
		*cq++ = Base64[a];
		*cq++ = Base64[b];
		*cq++ = '=';
		*cq = '=';
	} else if (remlen==2) {
		a=(cp[0]>>2);
		b=(cp[0]<<4)&0x30;
		b|=(cp[1]>>4);
		c=(cp[1]<<2)&0x3c;
		*cq++ = Base64[a];
		*cq++ = Base64[b];
		*cq++ = Base64[c];
		*cq = '=';
	}
	snprintf(dest, sizeof(dest), "=?%s?B?%s?=", MIME_CHARSET, encoded);
	return dest;
}

#ifdef SUPPORT_JIS
int jis2euc(char *str, char *buf, int buflen)
{
	char *str_tail;
	char *src, *dest;
	int len;
	int kanji_in = 0;

 	len = strlen(str);
	src = str;
	dest = buf;
	str_tail = src + len;
	while (src < str_tail && dest < (buf + buflen)) {
		/*  
		   toggle state.
		   ESC $ B ===> KANJI IN
		   ESC ( B ===> KANJI OUT
		*/
		if (*src == 0x1b) {
			if (src[1] == '$' && src[2] == 'B') {
				kanji_in = 1;
				src += 3;
				continue;
			} else if (src[1] == '(' && src[2] == 'B') {
				kanji_in = 0;
				src += 3;
				continue;
			}
		}
		if (kanji_in && 0x21 <= *src && *src <= 0x7e) {
			*dest++ = *src++ | 0x80;
		} else {
			*dest++ = *src++;
		}
	}
	*dest = '\0';
	return dest - buf;
}

int euc2jis(char *str, char *buf, int buflen)
{
	char *src_tail, *dest_tail;
	char *src, *dest;
	int len;
	int kanji_in = 0;

	len = strlen(str);
	src = str;
	dest = buf;
	src_tail = src + len;
	dest_tail = dest + buflen - 1;
	memset(buf, 0, buflen);
	while (src <= src_tail && dest <= dest_tail) {
		if (0 <= *src && *src <= 0x7e) {
			if (kanji_in) {
				kanji_in = 0;
				if (dest + 3 > dest_tail)
					break;
				*dest++ = 0x1b;
				*dest++ = '(';
				*dest++ = 'B';
			} 
			*dest++ = *src++;
		} else {
			if (!kanji_in) {
				kanji_in = 1;
				if (dest + 3 > dest_tail)
					break;
				*dest++ = 0x1b;
				*dest++ = '$';
				*dest++ = 'B';
			}
			*dest++ = *src++ & 0x7f;
		}
	}
	return dest - buf;
}
#endif // SUPPORT_JIS
